############################################################################
#
#   RestClient.t
#
############################################################################
use strict;

use Test::More tests => 11;
use DE_EPAGES::Core::API::Script qw (RunScript);
use DE_EPAGES::Core::API::Error qw ( GetError ExistsError);

use DE_EPAGES::Core::API::Object::RestClient;
use DE_EPAGES::Core::API::EpagesConfig qw (EpagesConfig);
use LWP::UserAgent;
use Test::Mock::LWP::Dispatch; # mock LWP requests!
use HTTP::Headers;
use HTTP::Response;
use HTTP::Status qw ( :constants :is status_message );

use constant ALIAS => 'SelfTest';
use constant INI_SECTION => 'REST_SERVICE';
use constant BASEURL => 'http://localhost:1337/rs/';
use constant DEFAULT_URL => 'http://localhost:8088/rs/';
use constant BROKENURL => 'http://brockenhost:80/rs/';

sub Main() {

    local $| = 1;

    my $UserAgent = Prepare();
    Tests($UserAgent);

}

sub Prepare() {

    my $Ini = EpagesConfig();
    # We don't write to disc, but use the cache of EpagesConfig()
    $Ini->set(INI_SECTION, ALIAS, BASEURL);

    my $header  = HTTP::Headers->new( 'Content-Type' => 'text/plain', );
    my $content = 'Always look bright side of life...';

    my $UA = LWP::UserAgent->new();
    $UA->agent('ePages TEST-Client');
    $UA->timeout(1);
    $UA->map(BASEURL . ALIAS ,
        sub {
                return HTTP::Response->new( HTTP_OK, status_message(HTTP_OK),
                $header, $content )
            });
#    $UA->map(BROKENURL . ALIAS ,
#        sub {
#                return HTTP::Response->new( HTTP_NOT_FOUND, status_message(HTTP_NOT_FOUND),
#                $header, $content )
#            });
    return $UA;
}

sub GetNewClient(;$) {
    my ($UserAgent) = @_;

    return DE_EPAGES::Core::API::Object::RestClient->new(
            Alias => ALIAS,
            Header => { 'Content_Type' => 'text/html' },
            UserAgent => $UserAgent // undef, # if undef RestClient creates its own $UserAgent
            Timeout => 1,
        );
}

sub Tests($) {
    my ($UserAgent) =@_;
    TestConfig();
    TestMethods($UserAgent);
}

sub TestConfig {
    my $Ini = EpagesConfig();

    is($Ini->get(INI_SECTION, ALIAS), BASEURL, 'setup from ini');

    my $Client = GetNewClient();
    is($Client->getServiceURL(), BASEURL . ALIAS, 'use config from ini'); # Read from Ini
    $Ini->delete(INI_SECTION, ALIAS);

    $Client = GetNewClient();
    # Read from Constants.pm
    is($Client->getServiceURL(),DEFAULT_URL . ALIAS, 'use dafaults (Constants.pm)');

    # Read from enviroment
    {
        local $ENV{EPAGESJ_HOST} = '10.10.10.10';
        local $ENV{EPAGESJ_PORT} = 666; ## no critic (ValuesAndExpressions::ProhibitMagicNumbers)
        $Client = GetNewClient();
        is( $Client->getServiceURL(), 'http://10.10.10.10:666/rs/' . ALIAS, 'use defaults (enviroment EPAGESJ_XXX)' );
    }

    # Read from enviroment and Constants.pm
    {
        local $ENV{EPAGESJ_HOST} = '10.10.10.10';
        # Port number is taken from Constants.pm
        $Client = GetNewClient();
        is( $Client->getServiceURL(), 'http://10.10.10.10:8088/rs/' . ALIAS, 'use defaults (mixed enviroment/Constants.pm)' );
    }

}

sub TestMethods() {
    my ($UserAgent) = @_;
    my $Client = GetNewClient($UserAgent);

    $Client->set('BaseURL', BASEURL);
    isa_ok($Client, 'DE_EPAGES::Core::API::Object::RestClient', 'constructor');
    is($Client->getBaseURL(), BASEURL, 'getBaseURL');
    is($Client->getServiceURL(), BASEURL . ALIAS, 'getServiceURL');

    my $Response = $Client->sendRequest('{"test":"just a test"}');
    like($Response, qr|(\S+\s+){1,}|, 'valid response');

    $Client->set('BaseURL', BROKENURL);
    eval{$Client->sendRequest('{"test":"just a test"}');};
    if (ExistsError){
        my $Error = GetError();
        ok($Error, 'error if response is invalid');
        like($Error->longMessage(),qr|404|, 'contains error code 404');
    } else {
        fail('Error expected from sendRequest()');
    }

}

RunScript( Sub => sub { Main() } );
