use Test::More tests => 60;

use POSIX;
use strict;
use DE_EPAGES::Address::API::Address qw ( IsValidBankName );
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Customer::API::Shop qw ( GetNextCustomerNo GetCustomerFolder );
use DE_EPAGES::DataCache::API::Cache qw( UpdateLocalConnection );
use DE_EPAGES::Database::API::Connection qw( ConnectTestSite DisconnectSite GetCurrentDBHandle );
use DE_EPAGES::Object::API::Factory qw( LoadRootObject ExistsObjectByPath JoinObjectPath LoadObjectByPath);
use DE_EPAGES::Object::API::Table::Country qw ( GetPKeyCountryByCode2 );
use DE_EPAGES::Order::API::Basket;
use DE_EPAGES::Order::API::Shop qw ( GetNextOrderNo );
use DE_EPAGES::PaymentTypes::API::Attributes::Basket;
use DE_EPAGES::PaymentTypes::UI::Basket;

local $ENV{'EPAGES_TESTSITE'} = 'Store' unless $ENV{'EPAGES_TESTSITE'};
my $ROOTPATH = 'Test';

sub Main {

    ConnectTestSite();

    # Setup database; usually each db section should contain an "key" entry where the aes can be found, like:
    #
    #   [Connection]
    #   datasource=dbi:mysql
    #   login=usr_sitedb
    #   passwd=epages
    #   active=1
    #   Database=sitedb
    #   StoreType=Site
    #   key=hst58a42hvbt8671loegysq87n4mc7z1
    #
    # if not, we add one key here.
    my $dbi = GetCurrentDBHandle();
    $dbi->{'key'} = 'hst58a42hvbt8671loegysq87n4mc7z1' unless $dbi->{'key'};

    # create test classes
    ok( Init(), 'initialization' );

    my $Shop   = LoadObjectByPath(JoinObjectPath(['', $ROOTPATH, 'TestShop1' ]));
    ok( $Shop->id, 'create shop ' . $Shop->id );

    my $CustomerNo = GetNextCustomerNo($Shop);
    ok( length $CustomerNo >= 4, 'customer no. starts with 1000' );

    my $CustomerFolder = GetCustomerFolder($Shop);
    ok( !$CustomerFolder->existsChild($CustomerNo), 'customer no. does not exists yet' );

    my $Customer = $Shop->get('CustomerClass')->insertObject( {
            Parent => $CustomerFolder,
            Alias  => $CustomerNo,
        }
    );
    ok( $Customer->id, 'insert new customer' );
    ok( $Customer->alias eq $CustomerNo, 'Customer.Alias' );

    my $OrderNo = GetNextOrderNo($Shop);
    ok( length $OrderNo >= 4, 'order no. starts with 1000' );

    my $Order = $Customer->folder('Orders')->insertChild(
        'CustomerOrder', {
            'Alias'    => $OrderNo,
            'Shop'     => $Shop,
            'Customer' => $Customer,
        }
    );

    # add a line item container
    my $Container = $Order->insertChild(
        'LineItemContainer', {
            'Alias'      => 'LineItemContainer',
            'LanguageID' => $Shop->get('LanguageID'),
            'LocaleID'   => $Shop->get('LocaleID'),
            'CurrencyID' => $Shop->get('CurrencyID'),
            'TaxArea'    => $Shop->get('TaxMatrix')->defaultTaxArea,
            'TaxModel'   => $Shop->get('TaxModel'),
        }
    );
    $Order->set( { 'LineItemContainer' => $Container } );

    foreach my $TryNumber ( 1 .. 2 ) {
        my $LineItemPaymentDirectDebit = $Order->insertChild(
            'LineItemPaymentDirectDebit', {
                'Alias'             => '08/15 ' . $TryNumber,
                'BankAccountHolder' => 'Max Mustermann',
                'BankAccountNo' => '1234567890',
                'BankCode' => '10020055',
                'BankName' => 'Musterbank',
                'LineItemContainer' => $Container,
            }
        );

        # EPG-15717: don't get value from cache (for every second try)
        UpdateLocalConnection( $dbi->cacheKey ) if 0 == $TryNumber % 2;

        # Get methods should decrypt the data transparently.
        is( $LineItemPaymentDirectDebit->get('BankAccountNo'), '1234567890',
            "BankAccountNo $TryNumber" );
        is( $LineItemPaymentDirectDebit->get('BankName'), 'Musterbank', "BankName $TryNumber" );

        # Get the encrypted data from DB directly.
        my $SQL = 'SELECT bankaccountno, bankname FROM lineitempaymentdirectdebit WHERE lineitempaymentddbtid = ?';
        my $aaRows = GetCurrentDBHandle()->execute( $SQL, [ $LineItemPaymentDirectDebit->id ] );

        use constant SHORT_ENCRYPTED_LENGTH => 64;
        is( length $aaRows->[0]->[0], SHORT_ENCRYPTED_LENGTH, "BankAccountNo encrypted $TryNumber" );
        is( length $aaRows->[0]->[1], SHORT_ENCRYPTED_LENGTH, "BankName encrypted $TryNumber" );

        # Update test
        # Get methods should decrypt the data transparently.
        $LineItemPaymentDirectDebit->set({'BankAccountNo' => '0987654321'});
        is( $LineItemPaymentDirectDebit->get('BankAccountNo'), '0987654321',
            'BankAccountNo Update');

        $LineItemPaymentDirectDebit->set({'BankName' => 'Muster_Bank'});
        is( $LineItemPaymentDirectDebit->get('BankName'), 'Muster_Bank', 'BankName Update' );
    }

    # EPG-6785
    my $User = $Shop->insertChild( 'User', { Alias => 'TestUser1', Password=> '' } );
    $Shop->featureMaxValue('Basket', 1);
    my $Basket = DE_EPAGES::Order::API::Basket::GetBasket($User, $Shop);
    my $ShopAddress = $Order->insertChild( 'Address', {
        'Alias' => 'ShopAddress',
    });
    my $BillingAddress = $Order->insertChild( 'Address', {
        'Alias' => 'BillingAddress',
    });
    # test all non-mandatory cases
    foreach my $CountryA ( qw ( NL AN AW )) {
        $ShopAddress->set({ 'CountryID' => GetPKeyCountryByCode2($CountryA)});
        foreach my $CountryB ( qw ( NL AN AW )) {
            $BillingAddress->set({ 'CountryID' => GetPKeyCountryByCode2($CountryB)});
            $Basket->set({
               'ShopAddress'    => $ShopAddress,
               'BillingAddress' => $BillingAddress,
            });
            ## no critic qw(ProtectPrivateSubs)
            my $IsMandatory = DE_EPAGES::PaymentTypes::API::Attributes::Basket::_IsBankCodeMandatory($Basket);
            is ( $IsMandatory, 0, "Bank code is optional ($CountryA, $CountryB)");
        }
    }
    # test cases where the shop address equals a country where the bank code is mandatory
    foreach my $CountryA ( qw ( DE US AQ )) {
        $ShopAddress->set({ 'CountryID' => GetPKeyCountryByCode2($CountryA)});
        foreach my $CountryB ( qw ( NL AN AW )) {
            $BillingAddress->set({ 'CountryID' => GetPKeyCountryByCode2($CountryB)});
            $Basket->set({
               'ShopAddress'    => $ShopAddress,
               'BillingAddress' => $BillingAddress,
            });
            ## no critic qw(ProtectPrivateSubs)
            my $IsMandatory = DE_EPAGES::PaymentTypes::API::Attributes::Basket::_IsBankCodeMandatory($Basket);
            is ( $IsMandatory, 1, "Bank code is mandatory ($CountryA, $CountryB)");
        }
    }
    # test cases where the billing address equals a country where the bank code is mandatory
    foreach my $CountryA ( qw ( NL AN AW )) {
        $ShopAddress->set({ 'CountryID' => GetPKeyCountryByCode2($CountryA)});
        foreach my $CountryB ( qw ( DE US AQ )) {
            $BillingAddress->set({ 'CountryID' => GetPKeyCountryByCode2($CountryB)});
            $Basket->set({
               'ShopAddress'    => $ShopAddress,
               'BillingAddress' => $BillingAddress,
            });
            ## no critic qw(ProtectPrivateSubs)
            my $IsMandatory = DE_EPAGES::PaymentTypes::API::Attributes::Basket::_IsBankCodeMandatory($Basket);
            is ( $IsMandatory, 1, "Bank code is mandatory ($CountryA, $CountryB)");
        }
    }
    # test case for undefined shop address and billing address
    $Basket->set({ 'ShopAddress' => undef });
    ## no critic qw(ProtectPrivateSubs)
    my $IsMandatory = DE_EPAGES::PaymentTypes::API::Attributes::Basket::_IsBankCodeMandatory($Basket);
    is ( $IsMandatory, 1, 'Bank code is per default mandatory if the shop address is missing');
    $Basket->set({ 'ShopAddress' => $ShopAddress });
    $Basket->set({ 'BillingAddress' => undef });
    ## no critic qw(ProtectPrivateSubs)
    $IsMandatory = DE_EPAGES::PaymentTypes::API::Attributes::Basket::_IsBankCodeMandatory($Basket);
    is ( $IsMandatory, 1, 'Bank code is per default mandatory if the billing address is missing');
    $Basket->set({ 'ShopAddress' => undef });
    ## no critic qw(ProtectPrivateSubs)
    $IsMandatory = DE_EPAGES::PaymentTypes::API::Attributes::Basket::_IsBankCodeMandatory($Basket);
    is ( $IsMandatory, 1, 'Bank code is per default mandatory if both addresses are missing');

    # EPG-23503
    # List of bank names. \cf{http://en.wikipedia.org/wiki/List_of_banks_in_Europe}
    # test valid bank names
    my @aValidBankNames = (
        'Banque populaire de la Région économique de Strasbourg',
        'Caisse Centrale des Caisses d\'Epargne et de Prévoyanced',
        'Cassa di Risparmio delle Provincie Lombarde SPA (CARIPLO)',
        'Bank Delen & de Schaetzen NV Bank Delen of Bank de Schaetzen', # length 60
        'Ceskoslovenská obchodná banka',
        'Bank Sal.Oppenheim jr. & Cie (Österreich)',
        'Austria Wirtschaftsservice Gesellschaft (aws)',
        # although longer than 64 characters, the encrypted name fits in varchar(255)
        'Bank Delen & de Schaetzen NV Bank Delen of Bank de Schaetzen or "DDS Bank NV"',
    );
    foreach my $ValidBankName (@aValidBankNames) {
        my $IsValidBankName = IsValidBankName( $ValidBankName );
        is ( $IsValidBankName, 1, 'Bank name ' . substr($ValidBankName, 0, floor( (length $ValidBankName) / 2 )) .  '... is valid' );
    };
    # test invalid bank names
    my @aInvalidBankNames = (
        'BRE BANK S.A. (FORMERLY BANK ROZWOJU EKSPORTU S.A.) RETAIL BANKING Łódź, Al. Piłsudskiego 3, 90-368 Łódź', # too long
        # too many UTF-8 characters (length of encrypted hash gets too long)
        'ááááááááááééééééééééááááááááááééééééééééááááááááááééééééééééáááá',
    );
    foreach my $InvalidBankName (@aInvalidBankNames) {
        my $IsValidBankName = IsValidBankName( $InvalidBankName );
        is ( $IsValidBankName, 0, 'Bank name ' . substr($InvalidBankName, 0, floor( (length $InvalidBankName) / 3 )) .  '... is invalid' );
    }

    ok( Cleanup(), 'cleanup' );

    DisconnectSite();
}

sub Cleanup {
    return 1 if $ENV{'SKIP_CLEANUP'};
    LoadObjectByPath($ROOTPATH)->delete if ExistsObjectByPath($ROOTPATH);
    return 1;
}

sub Init {
    return 1 if LoadRootObject()->existsChild('Test') && $ENV{'SKIP_CLEANUP'};
    Cleanup();
    my $Folder = LoadRootObject()->insertChild( 'Object', { Alias => 'Test' } );
    $Folder->insertChild( 'Shop', { Alias => 'TestShop1' } );

    $Folder->insertChild(
        'Address', {
            'Alias'   => 'O1',
            'Street'  => 'Leutragraben',
            'City'    => 'Jena',
            'Zipcode' => '07740',
            'Company' => 'epages',
        }
    );

    return 1;
}

RunScript( Sub => \&Main, Debug => 1 );
